@@ -511,7 +511,7 @@ mod internal {
       /// the heap allocation that contains the `CancelHandle`. Without this
       /// extra weak reference, `Rc::get_mut()` might succeed and allow the
       /// `CancelHandle` to be moved when it isn't safe to do so.
-      weak_pin: Weak<dyn Any>,
+      _weak_pin: Weak<dyn Any>,
     },
     /// All item nodes in a chain are associated with a `Cancelable` head node.
     Item {
@@ -523,8 +523,8 @@ mod internal {
 
   impl NodeKind {
     fn head(rc_pin: &Rc<dyn Any>) -> Self {
-      let weak_pin = Rc::downgrade(rc_pin);
-      Self::Head { weak_pin }
+      let _weak_pin = Rc::downgrade(rc_pin);
+      Self::Head { _weak_pin }
     }
 
     fn item(waker: &Waker) -> Self {
